##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.feed.describeFeedGenerator`."""

    did: str  #: Did.
    feeds: t.List['models.AppBskyFeedDescribeFeedGenerator.Feed']  #: Feeds.
    links: t.Optional['models.AppBskyFeedDescribeFeedGenerator.Links'] = None  #: Links.


@dataclass
class Feed(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.describeFeedGenerator`."""

    uri: str  #: Uri.

    _type: str = 'app.bsky.feed.describeFeedGenerator#feed'


@dataclass
class Links(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.describeFeedGenerator`."""

    privacyPolicy: t.Optional[str] = None  #: Privacy policy.
    termsOfService: t.Optional[str] = None  #: Terms of service.

    _type: str = 'app.bsky.feed.describeFeedGenerator#links'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto import CID
from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.sync.subscribeRepos`."""

    cursor: t.Optional[int] = None  #: The last known event to backfill from.


@dataclass
class Commit(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`."""

    blobs: t.List[CID]  #: Blobs.
    blocks: t.Union[str, bytes]  #: CAR file containing relevant blocks.
    commit: CID  #: Commit.
    ops: t.List['models.ComAtprotoSyncSubscribeRepos.RepoOp']  #: Ops.
    rebase: bool  #: Rebase.
    repo: str  #: Repo.
    seq: int  #: Seq.
    time: str  #: Time.
    tooBig: bool  #: Too big.
    prev: t.Optional[CID] = None  #: Prev.

    _type: str = 'com.atproto.sync.subscribeRepos#commit'


@dataclass
class Handle(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`."""

    did: str  #: Did.
    handle: str  #: Handle.
    seq: int  #: Seq.
    time: str  #: Time.

    _type: str = 'com.atproto.sync.subscribeRepos#handle'


@dataclass
class Migrate(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`."""

    did: str  #: Did.
    seq: int  #: Seq.
    time: str  #: Time.
    migrateTo: t.Optional[str] = None  #: Migrate to.

    _type: str = 'com.atproto.sync.subscribeRepos#migrate'


@dataclass
class Tombstone(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`."""

    did: str  #: Did.
    seq: int  #: Seq.
    time: str  #: Time.

    _type: str = 'com.atproto.sync.subscribeRepos#tombstone'


@dataclass
class Info(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`."""

    name: str  #: Name.
    message: t.Optional[str] = None  #: Message.

    _type: str = 'com.atproto.sync.subscribeRepos#info'


@dataclass
class RepoOp(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`."""

    action: str  #: Action.
    path: str  #: Path.
    cid: t.Optional[CID] = None  #: Cid.

    _type: str = 'com.atproto.sync.subscribeRepos#repoOp'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client.models import base


@dataclass
class Label(base.ModelBase):

    """Definition model for :obj:`com.atproto.label.defs`. Metadata tag on an atproto resource (eg, repo or record)"""

    cts: str  #: timestamp when this label was created.
    src: str  #: DID of the actor who created this label.
    uri: str  #: AT URI of the record, repository (account), or other resource which this label applies to.
    val: str  #: the short string name of the value or type of this label.
    cid: t.Optional[
        str
    ] = None  #: optionally, CID specifying the specific version of 'uri' resource this label applies to.
    neg: t.Optional[bool] = None  #: if true, this is a negation label, overwriting a previous label.

    _type: str = 'com.atproto.label.defs#label'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.notification.listNotifications`."""

    cursor: t.Optional[str] = None  #: Cursor.
    limit: t.Optional[int] = None  #: Limit.
    seenAt: t.Optional[str] = None  #: Seen at.


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.notification.listNotifications`."""

    notifications: t.List['models.AppBskyNotificationListNotifications.Notification']  #: Notifications.
    cursor: t.Optional[str] = None  #: Cursor.


@dataclass
class Notification(base.ModelBase):

    """Definition model for :obj:`app.bsky.notification.listNotifications`."""

    author: 'models.AppBskyActorDefs.ProfileView'  #: Author.
    cid: str  #: Cid.
    indexedAt: str  #: Indexed at.
    isRead: bool  #: Is read.
    reason: str  #: Expected values are 'like', 'repost', 'follow', 'mention', 'reply', and 'quote'.
    record: 'base.RecordModelBase'  #: Record.
    uri: str  #: Uri.
    labels: t.Optional[t.List['models.ComAtprotoLabelDefs.Label']] = None  #: Labels.
    reasonSubject: t.Optional[str] = None  #: Reason subject.

    _type: str = 'app.bsky.notification.listNotifications#notification'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.repo.listRecords`."""

    collection: str  #: The NSID of the record type.
    repo: str  #: The handle or DID of the repo.
    cursor: t.Optional[str] = None  #: Cursor.
    limit: t.Optional[int] = None  #: The number of records to return.
    reverse: t.Optional[bool] = None  #: Reverse the order of the returned records?
    rkeyEnd: t.Optional[str] = None  #: DEPRECATED: The highest sort-ordered rkey to stop at (exclusive).
    rkeyStart: t.Optional[str] = None  #: DEPRECATED: The lowest sort-ordered rkey to start from (exclusive).


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.repo.listRecords`."""

    records: t.List['models.ComAtprotoRepoListRecords.Record']  #: Records.
    cursor: t.Optional[str] = None  #: Cursor.


@dataclass
class Record(base.ModelBase):

    """Definition model for :obj:`com.atproto.repo.listRecords`."""

    cid: str  #: Cid.
    uri: str  #: Uri.
    value: 'base.RecordModelBase'  #: Value.

    _type: str = 'com.atproto.repo.listRecords#record'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.listAppPasswords`."""

    passwords: t.List['models.ComAtprotoServerListAppPasswords.AppPassword']  #: Passwords.


@dataclass
class AppPassword(base.ModelBase):

    """Definition model for :obj:`com.atproto.server.listAppPasswords`."""

    createdAt: str  #: Created at.
    name: str  #: Name.

    _type: str = 'com.atproto.server.listAppPasswords#appPassword'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.admin.getRecord`."""

    uri: str  #: Uri.
    cid: t.Optional[str] = None  #: Cid.


#: Response reference to :obj:`models.ComAtprotoAdminDefs.RecordViewDetail` model.
ResponseRef: t.Type[models.ComAtprotoAdminDefs.RecordViewDetail] = models.ComAtprotoAdminDefs.RecordViewDetail



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.sync.getRepo`."""

    did: str  #: The DID of the repo.
    earliest: t.Optional[str] = None  #: The earliest commit in the commit range (not inclusive).
    latest: t.Optional[str] = None  #: The latest commit in the commit range (inclusive).


#: Response raw data type.
Response: t.Union[t.Type[str], t.Type[bytes]] = bytes



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.label.subscribeLabels`."""

    cursor: t.Optional[int] = None  #: The last known event to backfill from.


@dataclass
class Labels(base.ModelBase):

    """Definition model for :obj:`com.atproto.label.subscribeLabels`."""

    labels: t.List['models.ComAtprotoLabelDefs.Label']  #: Labels.
    seq: int  #: Seq.

    _type: str = 'com.atproto.label.subscribeLabels#labels'


@dataclass
class Info(base.ModelBase):

    """Definition model for :obj:`com.atproto.label.subscribeLabels`."""

    name: str  #: Name.
    message: t.Optional[str] = None  #: Message.

    _type: str = 'com.atproto.label.subscribeLabels#info'



from io import BytesIO
from typing import List, Union

import dag_cbor as _dag_cbor

DagCborData = Union[bytes, bytearray, BytesIO]


class _BytesReadCounter:
    _num_bytes_read = 0

    def __call__(self, _, num_bytes_read: int) -> None:
        self._num_bytes_read += num_bytes_read

    def __int__(self) -> int:
        return self._num_bytes_read


def decode_dag(data: DagCborData, *, allow_concat: bool = False, callback=None) -> _dag_cbor.IPLDKind:
    """Decodes and returns a single data item from the given data, with the DAG-CBOR codec.

    Args:
        data: The stream of bytes or bytes with DAG-CBOR data.
        allow_concat: Allow partial stream decoding.
        callback: Callback on every time an item is decoded.

    Returns:
        :obj:`dag_cbor.IPLDKind`: Decoded DAG-CBOR.
    """
    return _dag_cbor.decode(data, allow_concat=allow_concat, callback=callback)


def decode_dag_multi(data: DagCborData) -> List[_dag_cbor.IPLDKind]:
    """Decodes and returns many data items from the given data, with the DAG-CBOR codec.

    Args:
        data: The stream of bytes or bytes with DAG-CBOR data.

    Returns:
        :obj:`list` of :obj:`dag_cbor.IPLDKind`: Decoded DAG-CBOR.
    """

    if not isinstance(data, BytesIO):
        data = BytesIO(data)

    data_size = data.getbuffer().nbytes

    data_parts = []
    if data_size == 0:
        return data_parts

    bytes_read_counter = _BytesReadCounter()
    while int(bytes_read_counter) != data_size:
        decoded_part = decode_dag(data, allow_concat=True, callback=bytes_read_counter)
        data_parts.append(decoded_part)

    return data_parts



import asyncio
import random
import threading
import time
import traceback
import typing as t

import httpx
from httpx_ws import (
    WebSocketDisconnect,
    WebSocketInvalidTypeReceived,
    WebSocketNetworkError,
    WebSocketUpgradeError,
    aconnect_ws,
    connect_ws,
)

from atproto.exceptions import FirehoseError
from atproto.firehose.models import Frame
from atproto.xrpc_client.models.common import XrpcError

if t.TYPE_CHECKING:
    from httpx_ws import AsyncWebSocketSession, WebSocketSession

    from atproto.firehose.models import MessageFrame

_BASE_WEBSOCKET_URL = 'https://bsky.social/xrpc'

OnMessageCallback = t.Callable[['MessageFrame'], None]
AsyncOnMessageCallback = t.Callable[['MessageFrame'], t.Awaitable[None]]

OnCallbackErrorCallback = t.Callable[[BaseException], None]


def _build_websocket_url(method: str, base_url: t.Optional[str] = None) -> str:
    if base_url is None:
        base_url = _BASE_WEBSOCKET_URL

    return f'{base_url}/{method}'


def _handle_firehose_error_or_stop(exception: Exception) -> bool:
    """Returns if the connection should be properly be closed or reraise exception."""

    if isinstance(exception, WebSocketDisconnect):
        if exception.code == 1000:
            return True
        if exception.code in {1001, 1002, 1003}:
            return False
    if isinstance(exception, WebSocketNetworkError):
        return False
    if isinstance(exception, httpx.NetworkError):
        return False
    if isinstance(exception, httpx.TimeoutException):
        return False
    if isinstance(exception, WebSocketInvalidTypeReceived):
        raise FirehoseError from exception
    if isinstance(exception, WebSocketUpgradeError):
        raise FirehoseError from exception
    if isinstance(exception, FirehoseError):
        raise exception

    raise FirehoseError from exception


class _WebsocketClientBase:
    def __init__(
        self,
        method: str,
        base_url: t.Optional[str] = None,
        params: t.Optional[t.Dict[str, t.Any]] = None,
        *,
        async_version: bool = False,
    ) -> None:
        self._params = params
        self._url = _build_websocket_url(method, base_url)

        self._async_version = async_version

        self._reconnect_no = 0
        self._max_reconnect_delay_sec = 64

        self._on_message_callback: t.Optional[t.Union[OnMessageCallback, AsyncOnMessageCallback]] = None
        self._on_callback_error_callback: t.Optional[OnCallbackErrorCallback] = None

    def _get_client(
        self,
    ) -> t.Union[t.AsyncGenerator['AsyncWebSocketSession', None], t.Generator['WebSocketSession', None, None]]:
        if self._async_version:
            return aconnect_ws(self._url, params=self._params)

        return connect_ws(self._url, params=self._params)

    def _get_reconnection_delay(self) -> int:
        base_sec = 2**self._reconnect_no
        rand_sec = random.uniform(-0.5, 0.5)  # noqa: S311

        return min(base_sec, self._max_reconnect_delay_sec) + rand_sec

    def _process_raw_frame(self, data: bytes) -> None:
        frame = Frame.from_bytes(data)
        if frame.is_error:
            raise FirehoseError(XrpcError(frame.body.error, frame.body.message))
        if frame.is_message:
            self._process_message_frame(frame)
        else:
            raise FirehoseError('Unknown frame type')

    def start(
        self,
        on_message_callback: t.Union[OnMessageCallback, AsyncOnMessageCallback],
        on_callback_error_callback: t.Optional[OnCallbackErrorCallback] = None,
    ) -> None:
        """Subscribe to Firehose and start client.

        Args:
            on_message_callback: Callback that will be called on the new Firehose message.
            on_callback_error_callback: Callback that will be called if the `on_message_callback` raised an exception.

        Returns:
            :obj:`None`
        """
        raise NotImplementedError

    def stop(self) -> None:
        """Unsubscribe and stop Firehose client.

        Returns:
            :obj:`None`
        """
        raise NotImplementedError

    def _process_message_frame(self, frame: 'MessageFrame') -> None:
        raise NotImplementedError


class _WebsocketClient(_WebsocketClientBase):
    def __init__(
        self, method: str, base_url: t.Optional[str] = None, params: t.Optional[t.Dict[str, t.Any]] = None
    ) -> None:
        super().__init__(method, base_url, params)

        self._stop_lock = threading.Lock()

    def _process_message_frame(self, frame: 'MessageFrame') -> None:
        try:
            self._on_message_callback(frame)
        except Exception as e:  # noqa: BLE001
            if self._on_callback_error_callback:
                try:
                    self._on_callback_error_callback(e)
                except:  # noqa
                    traceback.print_exc()
            else:
                traceback.print_exc()

    def start(
        self,
        on_message_callback: OnMessageCallback,
        on_callback_error_callback: t.Optional[OnCallbackErrorCallback] = None,
    ) -> None:
        self._on_message_callback = on_message_callback
        self._on_callback_error_callback = on_callback_error_callback

        while not self._stop_lock.locked():
            try:
                if self._reconnect_no != 0:
                    time.sleep(self._get_reconnection_delay())

                with self._get_client() as client:
                    self._reconnect_no = 0

                    while not self._stop_lock.locked():
                        raw_frame = client.receive_bytes()
                        self._process_raw_frame(raw_frame)
            except Exception as e:  # noqa: BLE001
                self._reconnect_no += 1

                should_stop = _handle_firehose_error_or_stop(e)
                if should_stop:
                    break

        self._stop_lock.release()

    def stop(self) -> None:
        if not self._stop_lock.locked():
            self._stop_lock.acquire()


class _AsyncWebsocketClient(_WebsocketClientBase):
    def __init__(
        self, method: str, base_url: t.Optional[str] = None, params: t.Optional[t.Dict[str, t.Any]] = None
    ) -> None:
        super().__init__(method, base_url, params, async_version=True)

        self._loop = asyncio.get_event_loop()
        self._on_message_tasks = set()

        self._stop_lock = asyncio.Lock()

    def _on_message_callback_done(self, task: asyncio.Task) -> None:
        self._on_message_tasks.discard(task)

        if task.exception():
            if not self._on_callback_error_callback:
                traceback.print_exc()
                return

            try:
                self._on_callback_error_callback(task.exception())
            except:  # noqa
                traceback.print_exc()

    def _process_message_frame(self, frame: 'MessageFrame') -> None:
        task = self._loop.create_task(self._on_message_callback(frame))
        self._on_message_tasks.add(task)
        task.add_done_callback(self._on_message_callback_done)

    async def start(
        self,
        on_message_callback: AsyncOnMessageCallback,
        on_callback_error_callback: t.Optional[OnCallbackErrorCallback] = None,
    ) -> None:
        self._on_message_callback = on_message_callback
        self._on_callback_error_callback = on_callback_error_callback

        while not self._stop_lock.locked():
            try:
                if self._reconnect_no != 0:
                    await asyncio.sleep(self._get_reconnection_delay())

                async with self._get_client() as client:
                    self._reconnect_no = 0

                    while not self._stop_lock.locked():
                        raw_frame = await client.receive_bytes()
                        self._process_raw_frame(raw_frame)
            except Exception as e:  # noqa: BLE001
                self._reconnect_no += 1

                should_stop = _handle_firehose_error_or_stop(e)
                if should_stop:
                    break

        self._stop_lock.release()

    async def stop(self) -> None:
        if not self._stop_lock.locked():
            await self._stop_lock.acquire()


FirehoseClient = _WebsocketClient
AsyncFirehoseClient = _AsyncWebsocketClient



import dataclasses
import json
import typing as t
from enum import Enum

from dacite import Config, exceptions, from_dict

from atproto.cid import CID
from atproto.exceptions import (
    MissingValueError,
    ModelError,
    ModelFieldError,
    UnexpectedFieldError,
    WrongTypeError,
)
from atproto.xrpc_client.models.base import ModelBase, RecordModelBase
from atproto.xrpc_client.models.blob_ref import BlobRef
from atproto.xrpc_client.models.type_conversion import RECORD_TYPE_TO_MODEL_CLASS

if t.TYPE_CHECKING:
    from atproto.xrpc_client.request import Response

M = t.TypeVar('M')


def _record_model_type_hook(data: dict) -> RecordModelBase:
    # used for inner Record types
    record_type = data.pop('$type')
    return get_or_create(data, RECORD_TYPE_TO_MODEL_CLASS[record_type])


def _decode_cid_hook(ref: t.Union[CID, str]) -> CID:
    if isinstance(ref, CID):
        return ref

    return CID.decode(ref)


_TYPE_HOOKS = {
    BlobRef: lambda ref: BlobRef.from_dict(ref),
    CID: _decode_cid_hook,
    RecordModelBase: _record_model_type_hook,
}
_DACITE_CONFIG = Config(cast=[Enum], type_hooks=_TYPE_HOOKS)


def get_or_create(
    model_data: t.Union[dict], model: t.Type[M] = None, *, strict: bool = True
) -> t.Optional[t.Union[M, dict]]:
    """Get model instance from raw data.

    Note:
        The record could have custom fields and be custom at all. For example, custom bsky clients add a "via" field
        to indicate that it was posted using not official client. Such custom types can't be decoded into proper models.
        You should work with it as with dict. By default, the method raises an exception on custom models.
        To fallback to dict type disable strict mode using the argument.

    Note:
        Auto-resolve of model works only with Record type for now.

    Args:
        model_data: Raw data.
        model: Class of the model or any another type. If None will be resolved automatically.
        strict: Fallback to raw data if can't properly parse.

    Returns:
        Instance of :obj:`model` or :obj:`None` or :obj:`dict` if `strict` is disabled.
    """
    try:
        return _get_or_create(model_data, model, strict=strict)
    except Exception as e:  # noqa: BLE001
        if strict:
            raise e

        return model_data


def _get_or_create(model_data: t.Union[dict], model: t.Type[M], *, strict: bool) -> t.Optional[t.Union[M, dict]]:
    if model_data is None:
        return None

    if model is None:
        # resolve model by $type and try to parse
        # resolves only Records
        record_type = model_data.pop('$type')
        if not record_type or record_type not in RECORD_TYPE_TO_MODEL_CLASS:
            return None

        return get_or_create(model_data, RECORD_TYPE_TO_MODEL_CLASS[record_type], strict=strict)

    if isinstance(model_data, model):
        return model_data

    try:
        # validate unexpected fields
        model(**model_data)

        return from_dict(model, model_data, config=_DACITE_CONFIG)
    except TypeError as e:
        # FIXME(MarshalX): "Params missing 1 required positional argument: 'rkey'" should raise another error
        msg = str(e).replace('__init__()', model.__name__)
        raise UnexpectedFieldError(msg) from e
    except exceptions.MissingValueError as e:
        raise MissingValueError(str(e)) from e
    except exceptions.WrongTypeError as e:
        raise WrongTypeError(str(e)) from e
    except exceptions.DaciteFieldError as e:
        raise ModelFieldError(str(e)) from e
    except exceptions.DaciteError as e:
        raise ModelError(str(e)) from e


def get_response_model(response: 'Response', model: t.Type[M]) -> t.Optional[M]:
    if model is bool:
        # Could not be False? Because the exception with errors will be raised from the server
        return response.success

    return get_or_create(response.content, model)


def _handle_dict_key(key: str) -> str:
    if key == '_type':  # System field. Replaced to original $ symbol because is not allowed in Python.
        return '$type'

    return key


def _handle_dict_value(ref: t.Any) -> t.Any:
    if isinstance(ref, BlobRef):
        return ref.to_dict()
    if isinstance(ref, CID):
        return ref.encode()

    return ref


def _model_as_dict_factory(value) -> dict:
    # exclude None values and process keys and values
    return {_handle_dict_key(k): _handle_dict_value(v) for k, v in value if v is not None}


def get_model_as_dict(model: t.Union[BlobRef, ModelBase]) -> dict:
    if model == BlobRef:
        return model.to_dict()

    if not dataclasses.is_dataclass(model):
        raise ModelError('Invalid model')

    return dataclasses.asdict(model, dict_factory=_model_as_dict_factory)


def get_model_as_json(model: t.Union[BlobRef, ModelBase]) -> str:
    return json.dumps(get_model_as_dict(model))


def is_json(json_data: t.Union[str, bytes]) -> bool:
    if isinstance(json_data, bytes):
        json_data.decode('UTF-8')

    try:
        json.loads(json_data)
        return True
    except:  # noqa
        return False


def is_record_type(model: t.Union[ModelBase, dict], types_module) -> bool:
    if isinstance(model, RecordModelBase) and hasattr(types_module, 'Main'):
        # for now records in Main. could be broken late
        if isinstance(model, dict):  # custom record
            return types_module.Main._type == model.get('$type')

        return types_module.Main._type == model._type

    return False



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Main(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.recordWithMedia`."""

    media: t.Union[
        'models.AppBskyEmbedImages.Main', 'models.AppBskyEmbedExternal.Main', 't.Dict[str, t.Any]'
    ]  #: Media.
    record: 'models.AppBskyEmbedRecord.Main'  #: Record.

    _type: str = 'app.bsky.embed.recordWithMedia'


@dataclass
class View(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.recordWithMedia`."""

    media: t.Union[
        'models.AppBskyEmbedImages.View', 'models.AppBskyEmbedExternal.View', 't.Dict[str, t.Any]'
    ]  #: Media.
    record: 'models.AppBskyEmbedRecord.View'  #: Record.

    _type: str = 'app.bsky.embed.recordWithMedia#view'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.actor.getProfile`."""

    actor: str  #: Actor.


#: Response reference to :obj:`models.AppBskyActorDefs.ProfileViewDetailed` model.
ResponseRef: t.Type[models.AppBskyActorDefs.ProfileViewDetailed] = models.AppBskyActorDefs.ProfileViewDetailed



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.feed.getPostThread`."""

    uri: str  #: Uri.
    depth: t.Optional[int] = None  #: Depth.
    parentHeight: t.Optional[int] = None  #: Parent height.


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.feed.getPostThread`."""

    thread: t.Union[
        'models.AppBskyFeedDefs.ThreadViewPost',
        'models.AppBskyFeedDefs.NotFoundPost',
        'models.AppBskyFeedDefs.BlockedPost',
        't.Dict[str, t.Any]',
    ]  #: Thread.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class ReplyRef(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.post`."""

    parent: 'models.ComAtprotoRepoStrongRef.Main'  #: Parent.
    root: 'models.ComAtprotoRepoStrongRef.Main'  #: Root.

    _type: str = 'app.bsky.feed.post#replyRef'


@dataclass
class Entity(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.post`. Deprecated: use facets instead."""

    index: 'models.AppBskyFeedPost.TextSlice'  #: Index.
    type: str  #: Expected values are 'mention' and 'link'.
    value: str  #: Value.

    _type: str = 'app.bsky.feed.post#entity'


@dataclass
class TextSlice(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.post`. Deprecated. Use app.bsky.richtext instead -- A text segment. Start is inclusive, end is exclusive. Indices are for utf16-encoded strings."""

    end: int  #: End.
    start: int  #: Start.

    _type: str = 'app.bsky.feed.post#textSlice'


@dataclass
class Main(base.RecordModelBase):

    """Record model for :obj:`app.bsky.feed.post`."""

    createdAt: str  #: Created at.
    text: str  #: Text.
    embed: t.Optional[
        t.Union[
            'models.AppBskyEmbedImages.Main',
            'models.AppBskyEmbedExternal.Main',
            'models.AppBskyEmbedRecord.Main',
            'models.AppBskyEmbedRecordWithMedia.Main',
            't.Dict[str, t.Any]',
        ]
    ] = None  #: Embed.
    entities: t.Optional[
        t.List['models.AppBskyFeedPost.Entity']
    ] = None  #: Deprecated: replaced by app.bsky.richtext.facet.
    facets: t.Optional[t.List['models.AppBskyRichtextFacet.Main']] = None  #: Facets.
    reply: t.Optional['models.AppBskyFeedPost.ReplyRef'] = None  #: Reply.

    _type: str = 'app.bsky.feed.post'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base
from atproto.xrpc_client.models.blob_ref import BlobRef


@dataclass
class Main(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`."""

    external: 'models.AppBskyEmbedExternal.External'  #: External.

    _type: str = 'app.bsky.embed.external'


@dataclass
class External(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`."""

    description: str  #: Description.
    title: str  #: Title.
    uri: str  #: Uri.
    thumb: t.Optional[BlobRef] = None  #: Thumb.

    _type: str = 'app.bsky.embed.external#external'


@dataclass
class View(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`."""

    external: 'models.AppBskyEmbedExternal.ViewExternal'  #: External.

    _type: str = 'app.bsky.embed.external#view'


@dataclass
class ViewExternal(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`."""

    description: str  #: Description.
    title: str  #: Title.
    uri: str  #: Uri.
    thumb: t.Optional[str] = None  #: Thumb.

    _type: str = 'app.bsky.embed.external#viewExternal'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.admin.resolveModerationReports`."""

    actionId: int  #: Action id.
    createdBy: str  #: Created by.
    reportIds: t.List[int]  #: Report ids.


#: Response reference to :obj:`models.ComAtprotoAdminDefs.ActionView` model.
ResponseRef: t.Type[models.ComAtprotoAdminDefs.ActionView] = models.ComAtprotoAdminDefs.ActionView



import typing as t
from dataclasses import dataclass

if t.TYPE_CHECKING:
    from atproto.xrpc_client.client.async_raw import AsyncClientRaw
    from atproto.xrpc_client.client.raw import ClientRaw


@dataclass
class NamespaceBase:
    _client: t.Union['ClientRaw', 'AsyncClientRaw']


@dataclass
class DefaultNamespace:
    """placeholder"""



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Main(base.ModelBase):

    """Definition model for :obj:`app.bsky.richtext.facet`."""

    features: t.List[
        t.Union['models.AppBskyRichtextFacet.Mention', 'models.AppBskyRichtextFacet.Link', 't.Dict[str, t.Any]']
    ]  #: Features.
    index: 'models.AppBskyRichtextFacet.ByteSlice'  #: Index.

    _type: str = 'app.bsky.richtext.facet'


@dataclass
class Mention(base.ModelBase):

    """Definition model for :obj:`app.bsky.richtext.facet`. A facet feature for actor mentions."""

    did: str  #: Did.

    _type: str = 'app.bsky.richtext.facet#mention'


@dataclass
class Link(base.ModelBase):

    """Definition model for :obj:`app.bsky.richtext.facet`. A facet feature for links."""

    uri: str  #: Uri.

    _type: str = 'app.bsky.richtext.facet#link'


@dataclass
class ByteSlice(base.ModelBase):

    """Definition model for :obj:`app.bsky.richtext.facet`. A text segment. Start is inclusive, end is exclusive. Indices are for utf8-encoded strings."""

    byteEnd: int  #: Byte end.
    byteStart: int  #: Byte start.

    _type: str = 'app.bsky.richtext.facet#byteSlice'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass

from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.sync.notifyOfUpdate`."""

    hostname: str  #: Hostname of the service that is notifying of update.



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class InviteCode(base.ModelBase):

    """Definition model for :obj:`com.atproto.server.defs`."""

    available: int  #: Available.
    code: str  #: Code.
    createdAt: str  #: Created at.
    createdBy: str  #: Created by.
    disabled: bool  #: Disabled.
    forAccount: str  #: For account.
    uses: t.List['models.ComAtprotoServerDefs.InviteCodeUse']  #: Uses.

    _type: str = 'com.atproto.server.defs#inviteCode'


@dataclass
class InviteCodeUse(base.ModelBase):

    """Definition model for :obj:`com.atproto.server.defs`."""

    usedAt: str  #: Used at.
    usedBy: str  #: Used by.

    _type: str = 'com.atproto.server.defs#inviteCodeUse'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


import typing as t
from dataclasses import dataclass

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Main(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.record`."""

    record: 'models.ComAtprotoRepoStrongRef.Main'  #: Record.

    _type: str = 'app.bsky.embed.record'


@dataclass
class View(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.record`."""

    record: t.Union[
        'models.AppBskyEmbedRecord.ViewRecord',
        'models.AppBskyEmbedRecord.ViewNotFound',
        'models.AppBskyEmbedRecord.ViewBlocked',
        'models.AppBskyFeedDefs.GeneratorView',
        't.Dict[str, t.Any]',
    ]  #: Record.

    _type: str = 'app.bsky.embed.record#view'


@dataclass
class ViewRecord(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.record`."""

    author: 'models.AppBskyActorDefs.ProfileViewBasic'  #: Author.
    cid: str  #: Cid.
    indexedAt: str  #: Indexed at.
    uri: str  #: Uri.
    value: 'base.RecordModelBase'  #: Value.
    embeds: t.Optional[
        t.List[
            t.Union[
                'models.AppBskyEmbedImages.View',
                'models.AppBskyEmbedExternal.View',
                'models.AppBskyEmbedRecord.View',
                'models.AppBskyEmbedRecordWithMedia.View',
                't.Dict[str, t.Any]',
            ]
        ]
    ] = None  #: Embeds.
    labels: t.Optional[t.List['models.ComAtprotoLabelDefs.Label']] = None  #: Labels.

    _type: str = 'app.bsky.embed.record#viewRecord'


@dataclass
class ViewNotFound(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.record`."""

    uri: str  #: Uri.

    _type: str = 'app.bsky.embed.record#viewNotFound'


@dataclass
class ViewBlocked(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.record`."""

    uri: str  #: Uri.

    _type: str = 'app.bsky.embed.record#viewBlocked'
